from airflow_commons.internal.athena.constants import EMPTY_STRING
from airflow_commons.internal.athena.query import get_add_partition_sql
from airflow_commons.internal.util.log_utils import get_logger

logger = get_logger("AthenaCore")


def add_partitions(
    client,
    result_bucket: str,
    result_folder: str,
    database: str,
    table: str,
    partitions: list,
):
    """
    Adds requested hive-like partitions to given Athena table
    :param client: Athena service
    :param result_bucket: Result bucket name
    :param result_folder: Result folder name
    :param database: Database name
    :param table: Table name
    :param partitions: Partitions dictionaries as key value pairs
    """
    sql = get_add_partition_sql(table=table, partitions=partitions)
    if sql == EMPTY_STRING:
        return
    return query(client, sql, result_bucket, result_folder, database)


def query(
    client,
    sql: str,
    result_bucket: str,
    result_folder: str,
    database: str,
):
    """
    Runs given query on Athena using AthenaClient
    :param client: Athena service
    :param sql: Query to run
    :param result_bucket: Result bucket name
    :param result_folder: Result folder name
    :param database: Database name
    """
    try:
        response = client.start_query_execution(
            QueryString=sql,
            QueryExecutionContext={"Database": database},
            ResultConfiguration={
                "OutputLocation": "s3://" + result_bucket + "/" + result_folder + "/"
            },
        )
    except Exception as e:
        logger.warn("An error occurred while running query")
        raise Exception(str(e)) from e
    return response


def get_query_execution(client, query_execution_id: str):
    """
    Returns current state of an athena query
    :param client: Athena service
    :param query_execution_id: Unique execution id of the query
    """
    return client.get_query_execution(QueryExecutionId=query_execution_id)



import argparse


class ParseKwargs(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        """
        Overrides default argparse argument parser action.
        """
        setattr(namespace, self.dest, dict())
        for value in values:
            key, value = value.split("=")
            getattr(namespace, self.dest)[key] = value



import requests
import threading
from datetime import datetime
from airflow_commons.internal.salesforce.constants import (
    DATA_PATH,
    UPSERT_ACCOUNT_PATH,
    UPDATE_ACCOUNT_PATH,
    QUERY_PATH,
)
from airflow_commons.internal.util.time_utils import get_interval_duration
from airflow_commons.internal.salesforce.auth import Connection, SalesCloudConnection
from airflow_commons.internal.salesforce.http_utils import (
    get_headers,
    get_async_operation_url,
    get_async_operation_result_url,
    get_async_operation_status_url,
    get_sync_operation_url,
    get_fetch_operation_url,
)
from airflow_commons.internal.salesforce.defaults import (
    TIMEOUT,
)
from airflow_commons.logger import get_logger


class SalesForceOperator(object):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        account_id: str,
        subdomain: str,
        scope: str = None,
        timeout: int = TIMEOUT,
    ):
        """
        Initializes SalesForceOperator instance with given parameters.
        If no scope is provided, connection starts with every operation available.

        :param client_id: Account client id string
        :param client_secret: Account client secret
        :param account_id: Account id
        :param subdomain: Account specific subdomain
        :param scope: Authorization scope, default value is None
        :param timeout: request timeout duration parameter
        """
        self.logger = get_logger("SalesForceOperator")
        self.connection = Connection(
            client_id=client_id,
            client_secret=client_secret,
            account_id=account_id,
            subdomain=subdomain,
            scope=scope,
        )
        self.session = requests.Session()
        self.timeout = timeout
        self.lock = threading.Lock()

    def async_upsert(self, key: str, data):
        """
        Starts an async upsert of given data to the data extension provided by key.
        :param key: External customer key of the data extension
        :param data: Data to be uploaded as json array
        :return:
        """
        return self._async_operation(key=key, data=data, method="PUT")

    def async_insert(self, key: str, data):
        """
        Starts an async insert of given data to the data extension provided by key.
        :param key: External customer key of the data extension
        :param data: Data to be uploaded as json array
        :return:
        """
        return self._async_operation(key=key, data=data, method="POST")

    def get_async_operation_result(self, request_id: str):
        """
        Returns result of an ongoing async process.
        :param request_id: ID of async request, can be found in the response of async request itself.
        :return:
        """
        url = get_async_operation_result_url(
            base_url=self.connection.rest_instance_url, request_id=request_id
        )
        request = requests.Request(
            method="GET",
            url=url,
            headers=get_headers(access_token=self.connection.access_token),
        )
        prep = request.prepare()
        return self.session.send(prep, timeout=self.timeout)

    def get_async_operation_status(self, request_id: str):
        """
        Returns current status of an ongoing async process.
        :param request_id: ID of async request, can be found in the response of async request itself.
        :return:
        """
        url = get_async_operation_status_url(
            base_url=self.connection.rest_instance_url, request_id=request_id
        )
        request = requests.Request(
            method="GET",
            url=url,
            headers=get_headers(access_token=self.connection.access_token),
        )
        prep = request.prepare()
        return self.session.send(prep, timeout=self.timeout)

    def _async_operation(self, key: str, data, method: str):
        """
        Internal method of async upsert and insert operations.
        :param key: External customer key of the data extension
        :param data: Data to be uploaded as json array
        :param method: HTTP method
        :return:
        """
        start = datetime.now()
        self.connection.check_and_refresh_token()
        url = get_async_operation_url(
            base_url=self.connection.rest_instance_url, key=key
        )
        request = requests.Request(
            method=method,
            url=url,
            json=data,
            headers=get_headers(access_token=self.connection.access_token),
        )
        prep = request.prepare()
        response = self.session.send(prep, timeout=self.timeout)
        try:
            response.raise_for_status()
            end = datetime.now()
            self.logger.debug(
                f"Async operation finished in {get_interval_duration(start, end)} seconds"
            )
        except Exception as e:
            if response:
                self.logger.error(
                    "An error occurred during async operation, "
                    + str(e)
                    + ", response from Salesforce "
                    + response.text
                )
        return response

    def sync_upsert(self, key: str, data):
        """
        Sync upsert of given data to the data extension provided by key.
        :param key: External customer key of the data extension
        :param data: Data to be uploaded as json array
        :return:
        """
        start = datetime.now()
        self.connection.check_and_refresh_token()
        url = get_sync_operation_url(
            base_url=self.connection.rest_instance_url, key=key
        )
        request = requests.Request(
            method="POST",
            url=url,
            json=data,
            headers=get_headers(access_token=self.connection.access_token),
        )
        prep = request.prepare()
        response = self.session.send(prep, timeout=self.timeout)
        try:
            response.raise_for_status()
            end = datetime.now()
            self.logger.debug(
                f"Sync operation finished in {get_interval_duration(start, end)} seconds"
            )
        except Exception as e:
            if response:
                self.logger.error(
                    "An error occurred during sync operation, "
                    + str(e)
                    + ", response from Salesforce "
                    + response.text
                )
        return response

    def sync_fetch(self, key: str, filter: str = None):
        """
        Sync upsert of given data to the data extension provided by key.
        :param key: External customer key of the data extension
        :param filter: Filter to apply to data extension search
        :return:
        """
        start = datetime.now()
        self.connection.check_and_refresh_token()
        url = get_fetch_operation_url(
            base_url=self.connection.rest_instance_url, key=key, filter=filter
        )
        request = requests.Request(
            method="GET",
            url=url,
            headers=get_headers(access_token=self.connection.access_token),
        )
        prep = request.prepare()
        response = self.session.send(prep, timeout=self.timeout)
        try:
            response.raise_for_status()
            end = datetime.now()
            self.logger.debug(
                f"Sync operation finished in {get_interval_duration(start, end)} seconds"
            )
        except Exception as e:
            if response:
                self.logger.error(
                    "An error occurred during sync operation, "
                    + str(e)
                    + ", response from Salesforce "
                    + response.text
                )
        return response

    def check_and_refresh(self):
        return self.connection.check_and_refresh_token()

    def get_next_page(self, next: str, timeout: int = TIMEOUT):
        """
        Creates request url and gets next page response
        :param next: Next page endpoint of previous page
        :param timeout: send request timeout duration in second, default is 30
        :return: requests.Response
        """
        self.connection.check_and_refresh_token()
        request_url = self.connection.rest_instance_url + DATA_PATH + next
        request = requests.Request(
            method="GET",
            url=request_url,
            headers=get_headers(access_token=self.connection.access_token),
        )
        return self.session.send(request.prepare(), timeout=timeout)


class SalesCloudOperator(object):
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        refresh_token: str,
        subdomain: str,
        scope: str = None,
        timeout: int = TIMEOUT,
    ):
        """
        Initializes SalesCloudOperator instance with given parameters.
        If no scope is provided, connection starts with every operation available.

        :param client_id: Account client id string
        :param client_secret: Account client secret
        :param refresh_token: Account refresh token
        :param subdomain: Authentication prefix for environment
        :param scope: Authorization scope, default value is None
        :param timeout: request timeout duration parameter
        """
        self.logger = get_logger("SalesCloudOperator")
        self.connection = SalesCloudConnection(
            client_id=client_id,
            client_secret=client_secret,
            refresh_token=refresh_token,
            subdomain=subdomain,
            scope=scope,
        )
        self.session = requests.Session()
        self.timeout = timeout

    def upsert_account(self, data):
        """
        Sync upsert of given data to the data extension provided by key.
        :param data: Data to be uploaded as json array
        :return:
        """
        start = datetime.now()
        self.connection.check_and_refresh_token()
        request = requests.Request(
            method="POST",
            url=self.connection.instance_url + UPSERT_ACCOUNT_PATH,
            json=data,
            headers=get_headers(access_token=self.connection.access_token),
        )
        prep = request.prepare()
        response = self.session.send(prep, timeout=self.timeout)
        try:
            response.raise_for_status()
            end = datetime.now()
            self.logger.debug(
                f"Sync operation finished in {get_interval_duration(start, end)} seconds"
            )
        except Exception as e:
            if response:
                self.logger.error(
                    "An error occurred during sync operation, "
                    + str(e)
                    + ", response from Salesforce "
                    + response.text
                )
        return response

    def update_account(self, data):
        """
        Sync update of given data to the data extension provided by key.
        :param data: Data to be uploaded as json array
        :return:
        """
        start = datetime.now()
        self.connection.check_and_refresh_token()
        request = requests.Request(
            method="POST",
            url=self.connection.instance_url + UPDATE_ACCOUNT_PATH,
            json=data,
            headers=get_headers(access_token=self.connection.access_token),
        )
        prep = request.prepare()
        response = self.session.send(prep, timeout=self.timeout)
        try:
            response.raise_for_status()
            end = datetime.now()
            self.logger.debug(
                f"Sync operation finished in {get_interval_duration(start, end)} seconds"
            )
        except Exception as e:
            if response:
                self.logger.error(
                    "An error occurred during sync operation, "
                    + str(e)
                    + ", response from Salesforce "
                    + response.text
                )
        return response

    def get_query_results(self, query):
        """
        Returns the results of the given query.
        :param query: SoqlQuery string
        :return:
        """
        start = datetime.now()
        self.connection.check_and_refresh_token()
        request = requests.Request(
            method="GET",
            url=self.connection.instance_url + QUERY_PATH + query,
            headers=get_headers(access_token=self.connection.access_token),
        )
        prep = request.prepare()
        response = self.session.send(prep, timeout=self.timeout)
        try:
            response.raise_for_status()
            end = datetime.now()
            self.logger.debug(
                f"Get SOQL query results finished in {get_interval_duration(start, end)} seconds"
            )
        except Exception as e:
            if response:
                self.logger.error(
                    "An error occurred during SOQL query operation, "
                    + str(e)
                    + ", response from Salesforce "
                    + response.text
                )
        return response



import json

from google.cloud import bigquery


def connect(service_account_file: str, service_account_str: str):
    """
    Connects to Bigquery client with given service account file path or string service account.

    :param service_account_file: relative location of service account json
    :param service_account_str: The string credential with a private key and other credentials information
    :return: a Client object instance required for API requests
    :raises: AssertionError: If there is no given service account
    """
    assert not (
        service_account_file is None and service_account_str is None
    ), "No service account has been provided"
    if service_account_str:
        info = json.loads(service_account_str)
        return bigquery.Client.from_service_account_info(info)
    return bigquery.Client.from_service_account_json(service_account_file)



ASYNC_PATH = "async"
ASYNC_JOB_RESULT_PATH = "/{request_id}/results"
ASYNC_JOB_STATUS_PATH = "/{request_id}/status"
AUTH_URL = "https://{subdomain}.auth.marketingcloudapis.com/"
CUSTOM_OBJECT_DATA_PATH = "/customobjectdata/key/{key}/rowset"
DATA_EXTENSION_KEY_PATH = "/dataextensions/key:{key}/rows"
DATA_EVENT_KEY_PATH = "/dataevents/key:{key}/rowset"
DATA_PATH = "data"
FILTER_QUERY = "$filter={filter}"
FORWARD_SLASH = "/"
GRANT_TYPE = "client_credentials"
QUERY_PATH = "/services/data/v64.0/query?q="
SERVICE_CLOUD_AUTH_URL = "https://{subdomain}.salesforce.com/services/oauth2"
SERVICE_CLOUD_GRANT_TYPE = "refresh_token"
SYNC_PATH = "hub"
TOKEN_PATH = "token"
UPDATE_ACCOUNT_PATH = "/services/apexrest/account/update"
UPSERT_ACCOUNT_PATH = "/services/apexrest/account/upsert"
V1 = "v1"
V2 = "v2"



import requests
from airflow_commons.internal.salesforce.http_utils import merge_path_parameters
from airflow_commons.internal.util.time_utils import datetime_add
from airflow_commons.logger import get_logger
from airflow_commons.internal.salesforce.constants import (
    GRANT_TYPE,
    AUTH_URL,
    TOKEN_PATH,
    V2,
    SERVICE_CLOUD_AUTH_URL,
    SERVICE_CLOUD_GRANT_TYPE,
    FORWARD_SLASH,
)
from datetime import datetime

from airflow_commons.resources.glossary import QUESTION_MARK


class Connection:
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        account_id: str,
        subdomain: str,
        scope: str = None,
    ):
        """
        Creates a connection instance to salesforce marketing cloud API

        :param client_id: Account client id string
        :param client_secret: Account client secret
        :param account_id: Account id
        :param subdomain: Account specific subdomain
        :param scope: Authorization scope
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.account_id = account_id
        self.auth_url = AUTH_URL.format(subdomain=subdomain)
        self.scope = scope
        self.logger = get_logger("Connection")
        self._authenticate()

    def _authenticate(self):
        """
        Authenticates to salesforce marketing cloud
        :return:
        """
        body = {
            "grant_type": GRANT_TYPE,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "account_id": self.account_id,
        }
        if self.scope is not None:
            body["scope"] = self.scope
        now = datetime.now()
        auth_response = requests.post(
            url=(self.auth_url + merge_path_parameters(V2, TOKEN_PATH)), json=body
        )
        try:
            auth_response.raise_for_status()
        except Exception as e:
            if auth_response:
                self.logger.error(
                    "An error occurred during authentication " + auth_response.text
                )
            raise Exception(str(e)) from e
        auth_response_json = auth_response.json()
        self.access_token = auth_response_json["access_token"]
        self.token_expiration_period = auth_response_json["expires_in"] - 60
        self.soap_instance_url = auth_response_json["soap_instance_url"]
        self.rest_instance_url = auth_response_json["rest_instance_url"]
        self.token_expiration_at = datetime_add(
            now, seconds=self.token_expiration_period
        )

    def check_and_refresh_token(self):
        """
        Checks if current token is expired, if so re-authenticates itself.
        :return:
        """
        now = datetime.now()
        if now > self.token_expiration_at:
            return
        self._authenticate()


class SalesCloudConnection:
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        refresh_token: str,
        subdomain: str,
        scope: str = None,
    ):
        """
        Creates a connection instance to salesforce service cloud API

        :param client_id: Account client id string
        :param client_secret: Account client secret
        :param subdomain: Authentication prefix for environment
        :param refresh_token: Refresh token
        :param scope: Authorization scope
        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.refresh_token = refresh_token
        self.auth_url = SERVICE_CLOUD_AUTH_URL.format(subdomain=subdomain)
        self.scope = scope
        self.logger = get_logger("Connection")
        self._authenticate()

    def _authenticate(self):
        """
        Authenticates to salesforce sales cloud
        :return:
        """
        body = {
            "grant_type": SERVICE_CLOUD_GRANT_TYPE,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }
        if self.scope is not None:
            body["scope"] = self.scope
        auth_url_extension = (
            f"grant_type={SERVICE_CLOUD_GRANT_TYPE}&client_id={self.client_id}"
            f"&client_secret={self.client_secret}&refresh_token={self.refresh_token}"
        )
        auth_response = requests.post(
            url=(
                self.auth_url
                + FORWARD_SLASH
                + TOKEN_PATH
                + QUESTION_MARK
                + auth_url_extension
            )
        )
        try:
            auth_response.raise_for_status()
        except Exception as e:
            if auth_response:
                self.logger.error(
                    "An error occurred during authentication " + auth_response.text
                )
            raise Exception(str(e)) from e

        auth_response_json = auth_response.json()
        self.access_token = auth_response_json["access_token"]
        self.signature = auth_response_json["signature"]
        self.scope = auth_response_json["scope"]
        self.instance_url = auth_response_json["instance_url"]
        self.id = auth_response_json["id"]
        self.token_type = auth_response_json["token_type"]
        self.token_expiration_at = datetime_add(
            datetime.fromtimestamp(int(auth_response_json["issued_at"]) / 1000),
            minutes=25,
        )

    def check_and_refresh_token(self):
        """
        Checks if current token is expired, if so re-authenticates itself.
        :return:
        """
        now = datetime.now()
        if now > self.token_expiration_at:
            return
        self._authenticate()



import sqlalchemy
from airflow_commons.logger import get_logger
from sqlalchemy import text, column
from sqlalchemy import update as sql_update
from sqlalchemy.dialects.mysql import insert


logger = get_logger("MySqlCore")


def get_table_metadata(table_name: str, engine_connection: sqlalchemy.engine.Engine):
    """
    Gets the metadata for the table provided with table_name.
    :param table_name: table name to get metadata
    :param engine_connection: database engine connection
    :return: sqlalchemy table.
    """
    return sqlalchemy.Table(
        table_name,
        sqlalchemy.MetaData(),
        autoload=True,
        autoload_with=engine_connection,
    )


def upsert(values: dict, session, table_name: str, update_column_names: list = None):
    """
    :param values: values to write into database
    :param session: session context manager
    :param table_name: database table name to write
    :param update_column_names: columns to be updated, default is all columns except the primary key
    :return:
    """
    table_metadata = get_table_metadata(
        table_name=table_name, engine_connection=session.bind
    )
    update_cols = {}
    insert_statement = insert(table_metadata).values(values)

    if update_column_names is None:
        update_column_names = []
        for name, col in insert_statement.table.columns.items():
            update_column_names.append(name)

    for col_name in update_column_names:
        if col_name not in table_metadata.primary_key:
            update_cols.update({col_name: insert_statement.inserted[col_name]})
    upsert_statement = insert_statement.on_duplicate_key_update(**update_cols)
    try:
        session.execute(upsert_statement)
    except Exception as e:
        logger.error(f"An exception occurred while upsert: {e}")
        raise


def update(table_name: str, values: list, session):
    """
    :param table_name: database table name to write
    :param values: list of values as dictionary with the optional where statement as a text
    :param session: session object
    """
    with session.begin() as current_session:
        table_metadata = get_table_metadata(
            table_name=table_name, engine_connection=current_session.bind
        )
        try:
            for statement in values:
                try:
                    where_clause = text(statement["where"])
                except KeyError:
                    where_clause = text("1=1")

                values_formatted = {}
                [
                    values_formatted.update({column(item["column"]): item["value"]})
                    for item in statement["values"]
                ]
                current_session.execute(
                    sql_update(table_metadata)
                    .where(where_clause)
                    .values(values_formatted)
                )
            current_session.commit()

        except Exception as e:
            logger.error(
                f"An exception {e} occurred, transaction rolled back. Nothing is changed."
            )
            current_session.rollback()
            raise e

        logger.debug("Row columns are updated successfully.")



import logging

LOG_LEVEL_MAP = {"info": logging.INFO, "debug": logging.DEBUG, "error": logging.ERROR}


def get_logger(name: str, log_level: str = "info"):
    logging.basicConfig(
        level=LOG_LEVEL_MAP[log_level],
        format="[%(asctime)s] {{%(filename)s:%(lineno)d}} %(levelname)s - %(message)s",
        force=True,
    )
    return logging.getLogger(name)



class ServiceAccount(object):
    def __init__(self, username: str = None, secret: str = None):
        """
        Constructs a new mixpanel service account instance.

        :param username: Service account username
        :param secret: Service account secret
        """
        self.username = username
        self.secret = secret

    def get_credentials(self):
        """
        Returns preserved credentials as basic auth parameters for this service account.

        :return: Tuple of service account credentials
        """
        creds = (self.username, self.secret)
        return creds


class Connection(object):
    def __init__(
        self,
        service_account_name: str = None,
        service_account_secret: str = None,
        access_token: str = None,
    ):
        """
        Constructs a new mixpanel connection instance.
        Supports both service account and token based authentications for various api call needs.

        :param service_account_name: Service account username
        :param service_account_secret: Service account secret
        :param access_token: Access token
        """
        self.service_account = ServiceAccount(
            username=service_account_name, secret=service_account_secret
        )
        self.access_token = access_token

    def get_service_account_credentials(self):
        """
        Should be used for basic auth based on service account credentials.

        :return: Tuple of service account credentials
        """
        return self.service_account.get_credentials()

    def get_access_token(self):
        """
        Should be used for token based auth

        :return: Access token
        """
        return self.access_token



from airflow_commons.logger import get_logger

from airflow_commons.internal.util.file_utils import read_sql

from airflow_commons.internal.mysql.auth import get_session
from airflow_commons.internal.mysql.core import upsert
from airflow_commons.internal.mysql.core import update as update_util
from airflow_commons.internal.mysql.query_utils import get_delete_sql
from airflow_commons.internal.mysql.query_utils import get_select_all_sql
from airflow_commons.internal.mysql.query_utils import get_select_sql
from airflow_commons.internal.mysql.query_utils import get_row_count_sql
import pandas as pd


class MysqlOperator(object):
    def __init__(self, username, password, host, db_name):
        """
        Initializes a MysqlOperator instance

        :param username: database username
        :param password: database password
        :param host: database host
        :param db_name: database name
        :return:
        """
        self.Session = get_session(username, password, host, db_name)
        self.logger = get_logger("MysqlOperator")

    def write_to_mysql(
        self,
        values: dict,
        chunk_size: int,
        table_name: str,
        update_column_names: list = None,
    ):
        """
        :param values: values to write into database
        :param chunk_size: data size to upload at a time
        :param table_name: database table name to write
        :param update_column_names: columns to be updated, default is all columns except the primary key
        :return:
        """

        chunks = [values[i : i + chunk_size] for i in range(0, len(values), chunk_size)]
        with self.Session.begin() as session:
            i = 0
            for chunk in chunks:
                i += 1
                upsert(
                    values=chunk,
                    session=session,
                    table_name=table_name,
                    update_column_names=update_column_names,
                )

    def delete(
        self,
        table_name: str,
        where_statement_file: str,
        where_statement_params: dict = None,
    ):
        """
        Runs a delete query on given table, and removes rows that conform where condition

        :param table_name: table name
        :param where_statement_file: relative location of where statement sql file
        :param where_statement_params: parameters of where statements
        """

        if where_statement_params is None:
            where_statement_params = dict()
        where_statement = read_sql(
            sql_file=where_statement_file, **where_statement_params
        )
        sql = get_delete_sql(
            table_name=table_name,
            where_statement=where_statement,
        )
        with self.Session.begin() as session:
            session.execute(sql)

    def select_all(
        self,
        table_name: str,
        where_statement_file: str,
        where_statement_params: dict = None,
        return_type: str = "resultProxy",
    ):
        """
        Runs a select query on given table and returns the rows that conform where condition

        :param table_name: database table name
        :param where_statement_params: relative location of where statement sql file
        :param where_statement_file: parameters of where statements
        :param return_type: parameter which determines return type
        :return: iterable ResultProxy object that stores results of the select query
        """

        if where_statement_params is None:
            where_statement_params = dict()
        where_statement = read_sql(
            sql_file=where_statement_file, **where_statement_params
        )
        sql = get_select_all_sql(
            table_name=table_name,
            where_statement=where_statement,
        )

        with self.Session.begin() as session:
            if return_type == "dataframe":
                result = pd.read_sql(sql, session.bind)
            else:
                result = session.execute(sql).fetchall()
        return result

    def select(
        self,
        table_name: str,
        column_names: list,
        where_statement_file: str,
        where_statement_params: dict = None,
        return_type: str = "resultProxy",
    ):
        """
        Runs a select query on given table and returns the rows that conform where condition

        :param table_name: database table name
        :param column_names: column name from the table
        :param where_statement_params: relative location of where statement sql file
        :param where_statement_file: parameters of where statements
        :param return_type: parameter which determines return type
        :return: iterable ResultProxy object that stores results of the select query
        """

        if where_statement_params is None:
            where_statement_params = dict()
        where_statement = read_sql(
            sql_file=where_statement_file, **where_statement_params
        )
        sql = get_select_sql(
            table_name=table_name,
            column_names=column_names,
            where_statement=where_statement,
        )
        with self.Session.begin() as session:
            if return_type == "dataframe":
                result = pd.read_sql(sql, session.bind)
            else:
                result = session.execute(sql).fetchall()

        return result

    def get_row_count(
        self,
        table_name: str,
        where_statement_file: str = None,
        where_statement_params: dict = None,
    ):
        if where_statement_file is not None:
            if where_statement_params is None:
                where_statement_params = dict()
            where_statement = read_sql(
                sql_file=where_statement_file, **where_statement_params
            )
        else:
            where_statement = ""

        sql = get_row_count_sql(
            table_name=table_name,
            where_statement=where_statement,
        )
        with self.Session.begin() as session:
            cursor = session.execute(sql)
            row_count = cursor.fetchone()[0]
        return row_count

    def update(
        self,
        table_name: str,
        values: list,
    ):
        """
        Runs an update query on given table, and updates row columns that conform where condition if given

        :param table_name: database table name to write
        :param values: list of values as dictionary with the optional where statement as a text
        exp: [{"values": [{"column": "column_name", "value": value_to_be_updated}],
            "where": "id = 1234"}]
        """

        update_util(table_name, values, self.Session)



BASE_URL = "https://api.mixpanel.com/"

ENGAGE_API_URL = BASE_URL + "engage?ip=0"
ENGAGE_BATCH_SIZE = 50

QUERY_START = "?"
AND = "&"
PROJECT_ID_QUERY_PARAM = "project_id="
STRICT = "strict=1"
IMPORT_API_URL = (
    BASE_URL
    + "import"
    + QUERY_START
    + STRICT
    + AND
    + PROJECT_ID_QUERY_PARAM
    + "{project_id}"
)

LOOKUP_TABLE_API_URL = BASE_URL + "lookup-tables"
GET_LOOKUP_TABLES = (
    LOOKUP_TABLE_API_URL + QUERY_START + PROJECT_ID_QUERY_PARAM + "{project_id}"
)
REPLACE_LOOKUP_TABLE = (
    LOOKUP_TABLE_API_URL
    + "/{table_id}"
    + QUERY_START
    + PROJECT_ID_QUERY_PARAM
    + "{project_id}"
)



from datetime import datetime
from typing import Union

import pandas as pd
import tempfile
from google.api_core import page_iterator
from google.cloud.bigquery import job

from airflow_commons.internal.bigquery.constants import STATE_FILTERS
from airflow_commons.resources.glossary import COMMA, WHITE_SPACE
from airflow_commons.logger import get_logger
from airflow_commons.resources.bigquery import (
    DEDUPLICATION_GET_AVAILABLE_PARTITIONS_FIELD_SQL_FILE,
    ARCHIVE_SOURCE_STATEMENT_FILE,
    ARCHIVE_DELETE_WHERE_STATEMENT_FILE,
)
from airflow_commons.internal.bigquery.auth import connect
from airflow_commons.internal.storage.auth import connect as storage_connect

from airflow_commons.internal.bigquery.defaults import (
    TIMEOUT,
    LOCATION,
    SELECT_RETURN_TYPE,
    STATE_FILTER,
)
from airflow_commons.internal.bigquery.core import (
    get_time_partition_field,
    single_value_select,
    query,
    select,
    query_information_schema,
    get_table_ref,
    get_primitive_column_list,
    get_job_config,
    extract,
)
from airflow_commons.internal.bigquery.query import (
    get_deduplication_source_statement,
    get_merge_sql,
    get_delete_sql,
    get_deduplication_additional_merge_conditions,
    get_truncate_sql,
)
from airflow_commons.internal.util.file_utils import read_sql
from airflow_commons.internal.util.time_utils import get_buffered_timestamp

logger = get_logger("gcp")


class BigqueryOperator(object):
    def __init__(
        self,
        service_account_file: str = None,
        service_account_str: str = None,
        enable_logging=False,
    ):
        """
        Initializes a BigqueryOperator instance
        :param service_account_file: relative location of service account file, by default is None
        :param enable_logging: boolean value for enabling to log queries, by default is false
        :param service_account_str: The string credential with a private key and other credentials information,
        by default is None
        :return:
        """
        self.ARCHIVE_MODES = ["INSERT", "UPSERT"]
        self.DEFAULT_PRIMARY_KEYS = ["id"]
        self.ARCHIVE_DEFAULT_PRIMARY_KEYS = ["id", "last_updated_at", "processed_at"]
        self.DEFAULT_TIME_COLUMNS = ["last_updated_at", "processed_at"]
        self.client = connect(service_account_file, service_account_str)
        self.logger = get_logger("BigqueryOperator")
        self.enable_logging = enable_logging

    def deduplicate(
        self,
        start_date: str,
        end_date: str,
        project_id: str,
        source_dataset: str,
        source_table: str,
        target_dataset: str,
        target_table: str,
        oldest_allowable_target_partition: str,
        primary_keys=None,
        time_columns=None,
        allow_partition_pruning: bool = True,
        job_priority: str = None,
        timeout=TIMEOUT,
        location=LOCATION,
    ):
        """
        Runs a merge query to deduplicate rows in historic table, and write to target snapshot table

        :param start_date: deduplication interval start
        :param end_date: deduplication interval end
        :param project_id: Bigquery project id
        :param source_dataset: source dataset id
        :param source_table: source table id
        :param target_dataset: target dataset id
        :param target_table: target table id
        :param oldest_allowable_target_partition: the oldest value of time partition column value to be added to target table, aims to keep partition count below limit
        :param primary_keys: primary key columns of the source and target tables
        :param time_columns: time columns list to order rows
        :param allow_partition_pruning: partition pruning allow parameter, if true prunes target table's partitions to limit query size
        :param job_priority: priority of bigquery job, it is currently BATCH or INTERACTIVE (default)
        :param timeout: query timeout duration parameter
        :param location: query location parameter
        :return:
        """
        job_config = get_job_config(job_priority)

        if primary_keys is None:
            primary_keys = self.DEFAULT_PRIMARY_KEYS
        if time_columns is None:
            time_columns = self.DEFAULT_TIME_COLUMNS
        buffered_start_date = get_buffered_timestamp(start_date)
        source_partition_field = get_time_partition_field(
            self.client, source_dataset, source_table
        )
        partition_pruning_params = dict()
        additional_merge_conditions = None
        partition_pruning_params["target_partition_field"] = get_time_partition_field(
            self.client, target_dataset, target_table
        )
        if allow_partition_pruning:
            partition_pruning_params["available_target_partitions"] = (
                single_value_select(
                    client=self.client,
                    sql=read_sql(
                        sql_file=DEDUPLICATION_GET_AVAILABLE_PARTITIONS_FIELD_SQL_FILE,
                        project_id=project_id,
                        source_dataset=source_dataset,
                        source_table=source_table,
                        target_partition_field=partition_pruning_params[
                            "target_partition_field"
                        ],
                        source_partition_field=source_partition_field,
                        start_date=buffered_start_date,
                        end_date=end_date,
                        oldest_allowable_target_partition=oldest_allowable_target_partition,
                    ),
                )
            )
            if partition_pruning_params["available_target_partitions"] is None:
                self.logger.info(
                    f"There is no data: - {project_id}.{source_dataset}.{source_table} - between [{start_date} - {end_date}] ."
                )
                return None
            additional_merge_conditions = get_deduplication_additional_merge_conditions(
                partition_pruning_params=partition_pruning_params
            )
        source_statement = get_deduplication_source_statement(
            start_date=buffered_start_date,
            end_date=end_date,
            project_id=project_id,
            source_dataset=source_dataset,
            source_table=source_table,
            source_partition_field=source_partition_field,
            primary_keys=primary_keys,
            time_columns=time_columns,
            oldest_allowable_target_partition=oldest_allowable_target_partition,
            allow_partition_pruning=allow_partition_pruning,
            partition_pruning_params=partition_pruning_params,
        )
        sql = get_merge_sql(
            client=self.client,
            project_id=project_id,
            target_dataset=target_dataset,
            target_table=target_table,
            source_statement=source_statement,
            primary_keys=primary_keys,
            mode="UPSERT",
            additional_merge_conditions=additional_merge_conditions,
        )
        return query(
            client=self.client,
            job_config=job_config,
            sql=sql,
            timeout=timeout,
            location=location,
            enable_logging=self.enable_logging,
        )

    def upsert(
        self,
        source_statement_file: str,
        project_id: str,
        target_dataset: str,
        target_table: str,
        source_statement_params: dict = None,
        primary_keys=None,
        additional_merge_conditions: str = None,
        job_priority: str = None,
        timeout=TIMEOUT,
        location=LOCATION,
    ):
        """
        Runs a merge query to upsert target table, with given source statement, merge conditions, and primary keys

        :param source_statement_file: relative location of source statement file
        :param project_id: Bigquery project id
        :param target_dataset: targeted dataset id
        :param target_table: targeted table id
        :param source_statement_params: parameters of source statement
        :param primary_keys: target table's primary key list
        :param additional_merge_conditions: additional user specified merge conditions as string
        :param job_priority: priority of bigquery job, it is currently BATCH or INTERACTIVE (default)
        :param timeout: query timeout duration in seconds
        :param location: query location parameter
        :return:
        """
        job_config = get_job_config(job_priority)

        if source_statement_params is None:
            source_statement_params = dict()
        if primary_keys is None:
            primary_keys = self.DEFAULT_PRIMARY_KEYS
        source_statement = read_sql(
            sql_file=source_statement_file, **source_statement_params
        )
        sql = get_merge_sql(
            client=self.client,
            project_id=project_id,
            target_dataset=target_dataset,
            target_table=target_table,
            source_statement=source_statement,
            primary_keys=primary_keys,
            mode="UPSERT",
            additional_merge_conditions=additional_merge_conditions,
        )
        return query(
            client=self.client,
            job_config=job_config,
            sql=sql,
            timeout=timeout,
            location=location,
            enable_logging=self.enable_logging,
        )

    def insert(
        self,
        source_statement_file: str,
        project_id: str,
        target_dataset: str,
        target_table: str,
        source_statement_params: dict = None,
        primary_keys=None,
        additional_merge_conditions: str = None,
        job_priority: str = None,
        timeout=TIMEOUT,
        location=LOCATION,
    ):
        """
        Runs a merge query to insert non-existing rows to target table, with given source statement, merge conditions, and primary keys

        :param source_statement_file: relative location of source statement file
        :param project_id: Bigquery project id
        :param target_dataset: targeted dataset id
        :param target_table: targeted table id
        :param source_statement_params: parameters of source statement
        :param additional_merge_conditions: additional user specified merge conditions as string
        :param primary_keys: target table's primary key list
        :param job_priority: priority of bigquery job, it is currently BATCH or INTERACTIVE (default)
        :param timeout: query timeout duration in seconds
        :param location: query location parameter
        :return:
        """
        job_config = get_job_config(job_priority)

        if source_statement_params is None:
            source_statement_params = dict()
        if primary_keys is None:
            primary_keys = self.DEFAULT_PRIMARY_KEYS
        source_statement = read_sql(
            sql_file=source_statement_file, **source_statement_params
        )
        sql = get_merge_sql(
            client=self.client,
            project_id=project_id,
            target_dataset=target_dataset,
            target_table=target_table,
            source_statement=source_statement,
            primary_keys=primary_keys,
            mode="INSERT",
            additional_merge_conditions=additional_merge_conditions,
        )
        return query(
            client=self.client,
            job_config=job_config,
            sql=sql,
            timeout=timeout,
            location=location,
            enable_logging=self.enable_logging,
        )

    def write_query_results_to_destination(
        self,
        sql_file: str,
        timeout: int,
        location: str,
        destination_dataset: str,
        destination_table: str,
        write_disposition: str,
        sql_params: dict = None,
        job_priority: str = None,
    ):
        """
        :param sql_file: relative location of sql file to read sql from
        :param timeout: job timeout parameter
        :param location: job location parameter
        :param destination_dataset: id of the destination dataset
        :param destination_table: id of the destination table
        :param write_disposition: write disposition of query results
        :param sql_params: sql parameters dictionary
        :param job_priority: priority of bigquery job, it is currently BATCH or INTERACTIVE (default)
        :return:
        """
        job_config = get_job_config(job_priority, write_disposition)
        job_config.allow_large_results = True
        job_config.destination = get_table_ref(
            self.client, destination_dataset, destination_table
        )

        if sql_params is None:
            sql_params = dict()
        sql = read_sql(sql_file=sql_file, **sql_params)
        return query(
            self.client,
            job_config,
            sql,
            timeout,
            location,
            self.enable_logging,
        )

    def get_query_results(
        self,
        sql_file: str,
        return_type: str = SELECT_RETURN_TYPE,
        sql_params: dict = None,
        timeout: int = TIMEOUT,
        location: str = LOCATION,
        index_label: str = None,
        job_priority: str = None,
        allow_large_results: bool = False,
    ):
        """
        Runs query and returns the result as dataframe

        :param return_type: file format to be returned (dict,dataframe,csv,json,json_string,pyarrow, rowIterator); default is dataframe
        :param sql_file: relative location of sql file to read sql from
        :param sql_params: sql parameters dictionary
        :param timeout: query timeout duration in seconds
        :param location: query location parameter
        :param index_label: if given and return type is csv, table will be indexed with this column name
        :param job_priority: priority of bigquery job, it is currently BATCH or INTERACTIVE (default)
        :param allow_large_results: allows large query results when true, default value is false
        :return: query result
        """
        if sql_params is None:
            sql_params = dict()
        sql = read_sql(sql_file=sql_file, **sql_params)
        return select(
            client=self.client,
            sql=sql,
            timeout=timeout,
            location=location,
            return_type=return_type,
            index_label=index_label,
            job_priority=job_priority,
            allow_large_results=allow_large_results,
        )

    def get_single_value_query_results(
        self,
        sql_file: str,
        sql_params: dict = None,
        job_priority: str = None,
        timeout=TIMEOUT,
        location=LOCATION,
    ):
        """
        Runs a single value returning query and returns the result

        :param sql_file: relative location of sql file to read sql from
        :param sql_params: sql parameters dictionary
        :param job_priority: priority of bigquery job, it is currently BATCH or INTERACTIVE (default)
        :param timeout: query timeout duration in seconds
        :param location: query location parameter
        :return: query result
        """
        if sql_params is None:
            sql_params = dict()
        sql = read_sql(sql_file=sql_file, **sql_params)
        return single_value_select(
            client=self.client,
            sql=sql,
            job_priority=job_priority,
            timeout=timeout,
            location=location,
        )

    def delete(
        self,
        project_id: str,
        dataset_id: str,
        table_id: str,
        where_statement_file: str,
        where_statement_params: dict = None,
        job_priority: str = None,
        timeout=TIMEOUT,
        location=LOCATION,
    ):
        """
        Runs a delete query on given table, and removes rows that conform where condition

        :param project_id: Bigquery project id
        :param dataset_id: dataset id
        :param table_id: table id
        :param where_statement_file: relative location of where statement sql file
        :param where_statement_params: parameters of where statements
        :param job_priority: priority of bigquery job, it is currently BATCH or INTERACTIVE (default)
        :param timeout: query timeout duration in seconds
        :param location: query location parameter
        :return: query result
        """
        job_config = get_job_config(job_priority)

        if where_statement_params is None:
            where_statement_params = dict()
        where_statement = read_sql(
            sql_file=where_statement_file, **where_statement_params
        )
        sql = get_delete_sql(
            project_id=project_id,
            dataset_id=dataset_id,
            table_id=table_id,
            where_statement=where_statement,
        )
        return query(
            client=self.client,
            job_config=job_config,
            sql=sql,
            timeout=timeout,
            location=location,
            enable_logging=self.enable_logging,
        )

    def truncate(
        self,
        project_id: str,
        dataset_id: str,
        table_id: str,
        job_priority: str = None,
        timeout=TIMEOUT,
        location=LOCATION,
    ):
        """
        Runs a truncate query on given table

        :param project_id: Bigquery project id
        :param dataset_id: dataset id
        :param table_id: table id
        :param job_priority: priority of bigquery job, it is currently BATCH or INTERACTIVE (default)
        :param timeout: query timeout duration in seconds
        :param location: query location parameter
        :return: query result
        """
        job_config = get_job_config(job_priority)

        sql = get_truncate_sql(
            project_id=project_id,
            dataset_id=dataset_id,
            table_id=table_id,
        )
        return query(
            client=self.client,
            job_config=job_config,
            sql=sql,
            timeout=timeout,
            location=location,
            enable_logging=self.enable_logging,
        )

    def archive(
        self,
        archive_date: str,
        project_id: str,
        source_dataset: str,
        source_table: str,
        target_dataset: str,
        target_table: str,
        primary_keys: list = None,
        mode: str = "INSERT",
        job_priority: str = None,
        timeout=TIMEOUT,
        location=LOCATION,
    ):
        """
        Archives selected rows on target non-partitioned table.

        :param archive_date: archive date
        :param project_id: Bigquery project id
        :param source_dataset: source dataset id
        :param source_table: source table id
        :param target_dataset: target dataset id
        :param target_table: target table id
        :param primary_keys: primary key columns of the source and target tables
        :param mode: archive mode, INSERT only inserts non-existing rows and UPSERT updates and inserts
        :param job_priority: priority of bigquery job, it is currently BATCH or INTERACTIVE (default)
        :param timeout: query timeout duration in seconds
        :param location: query location parameter
        :return:
        """
        if mode not in self.ARCHIVE_MODES:
            raise ValueError(
                "Invalid archive mode:{}. Acceptable column names are ".format(mode)
                + (COMMA + WHITE_SPACE).join(i for i in self.ARCHIVE_MODES)
            )

        if primary_keys is None or primary_keys == []:
            primary_keys = get_primitive_column_list(
                self.client, project_id, source_dataset, source_table
            )
        source_statement_params = dict()
        source_statement_params["project_id"] = project_id
        source_statement_params["source_dataset"] = source_dataset
        source_statement_params["source_table"] = source_table
        source_statement_params["archive_date"] = archive_date
        source_statement_params["source_partition_field"] = get_time_partition_field(
            client=self.client,
            dataset_id=source_dataset,
            table_id=source_table,
        )
        if mode == "INSERT":
            self.insert(
                source_statement_file=ARCHIVE_SOURCE_STATEMENT_FILE,
                project_id=project_id,
                target_dataset=target_dataset,
                target_table=target_table,
                source_statement_params=source_statement_params,
                primary_keys=primary_keys,
                job_priority=job_priority,
                timeout=timeout,
                location=location,
            )
        elif mode == "UPSERT":
            self.upsert(
                source_statement_file=ARCHIVE_SOURCE_STATEMENT_FILE,
                project_id=project_id,
                target_dataset=target_dataset,
                target_table=target_table,
                source_statement_params=source_statement_params,
                primary_keys=primary_keys,
                job_priority=job_priority,
                timeout=timeout,
                location=location,
            )
        delete_where_statement_params = dict()
        delete_where_statement_params["archive_date"] = archive_date
        delete_where_statement_params["source_partition_field"] = (
            source_statement_params["source_partition_field"]
        )
        self.delete(
            project_id=project_id,
            dataset_id=source_dataset,
            table_id=source_table,
            where_statement_file=ARCHIVE_DELETE_WHERE_STATEMENT_FILE,
            where_statement_params=delete_where_statement_params,
            job_priority=job_priority,
            timeout=timeout,
            location=location,
        )

    def list_tables(self, project_id: str, dataset_id: str):
        """
        Returns a list of all tables in a dataset

        :param project_id: Bigquery project id
        :param dataset_id: dataset id
        :return: List
        """
        return query_information_schema(
            client=self.client,
            requested_column_name="table_name",
            project_id=project_id,
            dataset_id=dataset_id,
        )

    def insert_to_table_from_dataframe(
        self, dataset_id: str, table_name: str, data: pd.DataFrame()
    ):
        """
        Inserts dataframe to the specified table. Returns list of errors for each row

        :param dataset_id: Bigquery dataset id
        :param table_name: Table name to insert rows
        :param data: Data to be inserted into the table
        :return: Sequence[Sequence[Mappings]]: A list with insert errors for each insert chunk.
        :raise: ValueError: if table's schema is not set
        """
        table_ref = get_table_ref(self.client, dataset_id, table_name)
        table = self.client.get_table(table_ref)
        return self.client.insert_rows_from_dataframe(table, data)

    def load_to_table_from_dataframe(
        self,
        dataset_id: str,
        table_name: str,
        write_disposition: str,
        data: pd.DataFrame(),
    ):
        """
        Loads dataframe to the specified table. Returns list of errors for each row

        :param dataset_id: Bigquery dataset id
        :param table_name: Table name to insert rows
        :param write_disposition: write disposition of query results
        :param data: Data to be inserted into the table
        :return This instance:
        :raises: ValueError: If a usable parquet engine cannot be found. This method requires :mod:`pyarrow` to be installed
        """
        job_config = get_job_config(
            job_priority="",
            write_disposition=write_disposition,
            job_config_str="LoadJobConfig",
        )

        table_ref = get_table_ref(self.client, dataset_id, table_name)
        return self.client.load_table_from_dataframe(
            data, table_ref, job_config=job_config
        )

    def load_to_table_from_json(
        self,
        dataset_id: str,
        table_name: str,
        write_disposition: str,
        data,
    ):
        """
        Loads json to the specified table. Returns list of errors for each row

        :param dataset_id: Bigquery dataset id
        :param table_name: Table name to insert rows
        :param write_disposition: write disposition of query results
        :param data: Data to be inserted into the table
        :return This instance:
        :raises: ValueError: If a usable parquet engine cannot be found. This method requires :mod:`pyarrow` to be installed
        """
        job_config = get_job_config(
            job_priority="",
            write_disposition=write_disposition,
            job_config_str="LoadJobConfig",
        )

        table_ref = get_table_ref(self.client, dataset_id, table_name)
        return self.client.load_table_from_json(
            data, table_ref, job_config=job_config
        ).result()

    def cancel_job(
        self, job_id: str
    ) -> Union[job.LoadJob, job.CopyJob, job.ExtractJob, job.QueryJob]:
        """
        Cancels given bigquery job.

        :param job_id: Job id to be cancelled
        :return: Union[
                google.cloud.bigquery.job.LoadJob,
                google.cloud.bigquery.job.CopyJob,
                google.cloud.bigquery.job.ExtractJob,
                google.cloud.bigquery.job.QueryJob,
            ]:
                Job instance, based on the resource returned by the API.
        """
        return self.client.cancel_job(job_id=job_id)

    def list_jobs(
        self,
        min_creation_time: datetime,
        max_creation_time: datetime,
        state_filter: str = STATE_FILTER,
        all_users: bool = True,
    ) -> page_iterator.Iterator:
        """
        Lists all jobs based on given state parameter and creation time intervals.

        :param min_creation_time:
        :param max_creation_time:
        :param state_filter: Job state filter. One of: "done", "pending", "running". Defaults to "running".
        :param all_users: If true, include jobs owned by all users in the project. Defaults to true.
        :return: google.api_core.page_iterator.Iterator: Iterable of job instances.
        :raises: ValueError: If on ``state_filter`` value is invalid
        """
        if state_filter not in STATE_FILTERS:
            raise ValueError(
                "Invalid state filter: {}. Acceptable values are {}".format(
                    state_filter, STATE_FILTERS
                )
            )
        return self.client.list_jobs(
            state_filter=state_filter,
            min_creation_time=min_creation_time,
            max_creation_time=max_creation_time,
            all_users=all_users,
        )

    def drop_table(self, table_id: str, not_found_ok: bool = False):
        """
        Deletes specified table from bigquery.

        :param table_id: Bigquery table id to delete.
        :param not_found_ok: If false, raises an exception when the table does not exist. Defaults to false.
        :raises: google.api_core.exceptions.NotFound: If given table id does not exist unless not_found_ok is true.
        :return:
        """
        self.client.delete_table(table_id, not_found_ok=not_found_ok)

    def extract_table_to_gcs(
        self,
        dataset_id: str,
        table_name: str,
        destination_uri: str,
        destination_format: str,
    ):
        """
        Extracts the content of given Bigquery table to GCS bucket.

        :param dataset_id: Bigquery source dataset
        :param table_name: Bigquery table to extracted
        :param destination_uri: GCS bucket uri to extract, wildcard uri
        :param destination_format: resulted file format
        :return: extract result
        """
        job_config = get_job_config(
            job_priority="",
            job_config_str="ExtractJobConfig",
            destination_format=destination_format,
        )

        table_ref = get_table_ref(self.client, dataset_id, table_name)
        return extract(
            client=self.client,
            job_config=job_config,
            source=table_ref,
            destination_uri=destination_uri,
            enable_logging=True,
        )


class StorageOperator(object):
    def __init__(
        self,
        service_account_file: str = None,
        service_account_str: str = None,
        enable_logging=False,
    ):
        """
        Initializes a BigqueryOperator instance
        :param service_account_file: relative location of service account file, by default is None
        :param enable_logging: boolean value for enabling to log queries, by default is false
        :param service_account_str: The string credential with a private key and other credentials information,
        by default is None
        :return:
        """
        self.client = storage_connect(service_account_file, service_account_str)
        self.logger = get_logger("StorageOperator")
        self.enable_logging = enable_logging

    def list_blobs(
        self,
        gcs_bucket_name: str,
        prefix: str,
        suffix: str,
    ):
        """
        List the file names in a GCS bucket

        :param gcs_bucket_name: Name of the GCS bucket
        :param prefix: prefix of the bucket containing table name
        :param suffix: suffix of the file represents the format e.g. parquet
        :return: List of file names
        """
        bucket = self.client.bucket(gcs_bucket_name)
        blobs = bucket.list_blobs(prefix=prefix)
        files = [blob.name for blob in blobs if blob.name.endswith(suffix)]
        return files

    def delete_blobs(
        self,
        gcs_bucket_name: str,
        file_names: list,
    ):
        """
        Deletes the given files from gcs bucket

        :param gcs_bucket_name: Name of the target bucket
        :param file_names: Name of the files to be deleted
        :return:
        """
        bucket = self.client.bucket(gcs_bucket_name)
        bucket.delete_blobs(file_names)

    def stream_parquet_file_from_gcs(self, gcs_bucket_name: str, file_names: list):
        """
        Generator method to download files from gcs

        :param gcs_bucket_name: Name of the GCS bucket
        :param file_names: List of the file names present in gcs
        :return: Single temporary file in memory each iteration
        """
        bucket = self.client.bucket(gcs_bucket_name)
        for file_name in file_names:
            blob = bucket.blob(file_name)
            with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                blob.download_to_file(temp_file)
                temp_file.flush()
                yield temp_file.name



import yaml


def read_config(config_file: str):
    """
    Reads config from a configuration file and returns a config dictionary

    :param config_file: relative location of config file
    :return: configs as dictionary
    """
    stream = open(config_file, "r")
    return yaml.safe_load(stream)



import json
from typing import List

import requests

from airflow_commons.internal.mixpanel.auth import Connection
from airflow_commons.internal.mixpanel.constants import (
    ENGAGE_API_URL,
    ENGAGE_BATCH_SIZE,
    IMPORT_API_URL,
    REPLACE_LOOKUP_TABLE,
    GET_LOOKUP_TABLES,
)
from airflow_commons.internal.mixpanel.defaults import TIMEOUT


class ServiceAccount(object):
    def __init__(self, username, secret):
        self.username = username
        self.secret = secret


class MixpanelOperator(object):
    def __init__(
        self,
        project_id: str = None,
        service_account_name: str = None,
        service_account_secret: str = None,
        access_token: str = None,
        timeout: int = TIMEOUT,
    ):
        self.session = requests.Session()
        self.project_id = project_id
        self.connection = Connection(
            service_account_name=service_account_name,
            service_account_secret=service_account_secret,
            access_token=access_token,
        )
        self.timeout = timeout

    def engage(self, values: List[dict]):
        """
        Updates profiles using engage API.
        :param values: List of profile updates. Must include $distinct_id and $set values. $set involves all profile parameters with their updated values.
        :return:
        """
        token = self.connection.get_access_token()
        for value in values:
            value["$token"] = token
        batches = [
            values[i : i + ENGAGE_BATCH_SIZE]
            for i in range(0, len(values), ENGAGE_BATCH_SIZE)
        ]
        for batch in batches:
            request = requests.Request(
                "POST",
                ENGAGE_API_URL,
                headers={"Content-Type": "application/json"},
                data=json.dumps(batch),
            )
            response = self._send_request(request)
            response.raise_for_status()

    def import_events(self, values: List[dict]):
        """
        Imports batch of values to Mixpanel.
        :param values: List of events to be sent. Each value should have keys event(str) and properties(dict).
        :return: Import API response.
        """
        url = IMPORT_API_URL.format(project_id=self.project_id)
        request = requests.Request(
            "POST",
            url,
            headers={"Accept": "application/json", "Content-Type": "application/json"},
            auth=self.connection.get_service_account_credentials(),
            data=json.dumps(values),
        )
        response = self._send_request(request)
        response.raise_for_status()
        return response

    def list_lookup_tables(self):
        """
        Lists available lookup tables in the project.
        :return: List of lookup table ids and names.
        """
        url = GET_LOOKUP_TABLES.format(project_id=self.project_id)
        request = requests.Request(
            "GET",
            url,
            headers={"Accept": "application/json"},
            auth=self.connection.get_service_account_credentials(),
        )
        response = self._send_request(request)
        response.raise_for_status()
        response_json = response.json()
        return response_json["results"]

    def replace_lookup_table(self, table_id: str, data: str):
        """
        Replaces existing lookup table with new data.
        :param table_id: Target lookup table id.
        :param data: Csv content of new data. First row must always be the header row. Also, first column is the id column.
        :return: Lookup API response.
        """
        url = REPLACE_LOOKUP_TABLE.format(table_id=table_id, project_id=self.project_id)
        request = requests.Request(
            "PUT",
            url,
            headers={"Accept": "application/json", "Content-Type": "text/csv"},
            auth=self.connection.get_service_account_credentials(),
            data=data,
        )
        response = self._send_request(request)
        response.raise_for_status()
        return response

    def _send_request(self, request: requests.Request):
        prep = request.prepare()
        return self.session.send(prep, timeout=self.timeout)
